<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.2/1705499
 */

namespace Reviews\Validator;

use Application\Validator\ConnectedAbstractValidator;
use Users\Validator\Users as UserValidator;
use Groups\Validator\Groups as GroupValidator;
use Groups\Model\Group;
use Reviews\UpdateService;
use Reviews\Model\Review as ReviewModel;

/**
 * Validator for the psuedo secret fields 'reviewers' and 'requiredReviewers' that handles
 * users and groups. Uses the existing Users and Groups validators.
 * @package Reviews\Validator
 */
class Reviewers extends ConnectedAbstractValidator
{
    private $connectionOption;
    private $projects;
    private $reviewAuthor;
    public $translator;

    public function __construct($p4, $reviewProjects, $reviewAuthor = null)
    {
        $this->connectionOption = array('connection' => $p4);
        $this->projects         = $reviewProjects;
        $this->reviewAuthor     = $reviewAuthor;
        $this->translator       = $p4->getService('translator');
        parent::__construct();
    }

    /**
     * Returns true if $value is an id for an existing user/group or if it contains a list of ids
     * representing existing users/groups in Perforce.
     *
     * @param   string|array    $value  id or list of ids to check
     * @return  boolean         true if value is id or list of ids of existing users/groups, false otherwise
     * @throws \Record\Exception\NotFoundException
     */
    public function isValid($value)
    {
        $retentionValid                    = true;
        $groupsValid                       = true;
        $usersValid                        = true;
        $quorumValid                       = true;
        $value                             = (array) $value;
        $ids                               = array();
        $this->abstractOptions['messages'] = array();

        // combined reviewers with the possibility of a quorum value being set for groups. Individuals will
        // have been in a requiredReviewers pseudo field implying true, we have built the combined field so
        // only need to validate that against retention policy
        foreach ($value as $key => $item) {
            $ids[] = $key;
            if (isset($item['required']) && !is_bool($item['required'])) {
                if (!is_numeric($item['required']) || (int) $item['required'] !== 1) {
                    array_push($this->abstractOptions['messages'], $this->translator->t("Quorum value must be 1"));

                    $quorumValid = false;
                }
            }
        }

        $users = array_filter(
            array_map(
                function ($id) {
                    return Group::isGroupName($id) === false ? $id : null;
                },
                $ids
            )
        );
        $groups = array_map(
            function ($id) {
                return Group::getGroupName($id);
            },
            array_diff($ids, $users)
        );

        if (count($groups)) {
            $groupsValid = $this->doValidate(new GroupValidator($this->connectionOption), $groups);
        }

        if (count($users)) {
            $usersValid = $this->doValidate(new UserValidator($this->connectionOption), $users);
        }

        $defaultRetainedReviewers = array();
        if ($this->projects !== null) {
            // everything seems valid so far - we need to check the reviewer retention policy on
            // associated projects and branches
            $defaultRetainedReviewers = UpdateService::mergeDefaultReviewersForProjects(
                $this->projects,
                $defaultRetainedReviewers,
                $this->connectionOption['connection'],
                array(UpdateService::ALWAYS_ADD_DEFAULT => false, UpdateService::FORCE_REQUIREMENT => false)
            );
        }
        // Exclude the author if they are default retained
        if ($this->reviewAuthor) {
            unset($defaultRetainedReviewers[$this->reviewAuthor]);
        }
        // For PHP backward compatibility
        $filter = $this;
        array_map(
            function ($key, $retention) use ($value, &$retentionValid, $filter) {
                if (isset($value[$key])) {
                    $minimum = $retention[ReviewModel::FIELD_MINIMUM_REQUIRED];
                    if ($minimum === true || $minimum === '1') {
                        $requiredValue = isset($value[$key]['required']) ? $value[$key]['required'] : null;
                        if ($requiredValue === null || ($requiredValue === '1' && $minimum === true)) {
                            $isGroup        = Group::isGroupName($key);
                            $requireMessage = $filter->translator->t('required');
                            $messageKey     = $key;
                            if ($isGroup) {
                                $messageKey     = Group::getGroupName($key);
                                $requireMessage = $minimum === true
                                    ? $filter->translator->t('require all') : $filter->translator->t('require 1');
                            }
                            $filter->addMessage(
                                $filter->translator->t(
                                    "'%s' must have a minimum requirement of '%s'",
                                    array($messageKey, $requireMessage)
                                )
                            );
                            $retentionValid = false;
                        }
                    }
                } else {
                    $filter->addMessage(
                        $filter->translator->t(
                            "'%s' must be a reviewer",
                            array(Group::isGroupName($key) ? Group::getGroupName($key) : $key)
                        )
                    );
                    $retentionValid = false;
                }
            },
            array_keys($defaultRetainedReviewers),
            $defaultRetainedReviewers
        );
        return $groupsValid && $usersValid && $quorumValid && $retentionValid;
    }

    /**
     * Runs the validation
     * @param $validator the validator
     * @param $value the value(s)
     * @return boolean true if value is valid
     */
    private function doValidate($validator, $value)
    {
        $valid = $validator->isValid($value);
        if (!$valid) {
            $this->abstractOptions['messages'] =
                array_merge($this->abstractOptions['messages'], $validator->getMessages());
        }
        return $valid;
    }
}
